# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose


import os
import unittest
from unittest.mock import patch

import pendulum
from azure.common import AzureMissingResourceHttpError
from azure.cosmosdb.table.tableservice import TableService
from azure.storage.blob import ContainerProperties

from mag_archiver.azure import create_table
from mag_archiver.mag import make_mag_query, MagState, MagDateType, MagRelease, MagTask, MagArchiverClient, \
    hide_if_not_none


class TestMag(unittest.TestCase):

    def test_hide_if_not_none(self):
        # Test that None is returned for None
        value = hide_if_not_none(None)
        self.assertEqual(value, None)

        # Test that 'hidden' is returned: string
        value = hide_if_not_none('hello world')
        self.assertEqual(value, 'hidden')

        # Test that 'hidden' is returned: integer
        value = hide_if_not_none(123)
        self.assertEqual(value, 'hidden')

    def test_make_mag_query(self):
        start_date = pendulum.datetime(year=2020, month=4, day=1)
        end_date = pendulum.datetime(year=2020, month=5, day=1)

        # No parameters
        query = make_mag_query()
        self.assertEqual(query, '')

        # State parameter
        query = make_mag_query(state=MagState.discovered)
        self.assertEqual(query, "State eq 'discovered'")

        query = make_mag_query(state=MagState.archived)
        self.assertEqual(query, "State eq 'archived'")

        query = make_mag_query(state=MagState.done)
        self.assertEqual(query, "State eq 'done'")

        # Start date parameter
        query = make_mag_query(start_date=start_date, date_type=MagDateType.release)
        self.assertEqual(query, "ReleaseDate ge datetime'2020-04-01T00:00Z'")

        query = make_mag_query(start_date=start_date, date_type=MagDateType.discovered)
        self.assertEqual(query, "DiscoveredDate ge datetime'2020-04-01T00:00Z'")

        query = make_mag_query(start_date=start_date, date_type=MagDateType.archived)
        self.assertEqual(query, "ArchivedDate ge datetime'2020-04-01T00:00Z'")

        query = make_mag_query(start_date=start_date, date_type=MagDateType.done)
        self.assertEqual(query, "DoneDate ge datetime'2020-04-01T00:00Z'")

        # End date parameter
        query = make_mag_query(end_date=end_date, date_type=MagDateType.release)
        self.assertEqual(query, "ReleaseDate lt datetime'2020-05-01T00:00Z'")

        query = make_mag_query(end_date=end_date, date_type=MagDateType.discovered)
        self.assertEqual(query, "DiscoveredDate lt datetime'2020-05-01T00:00Z'")

        query = make_mag_query(end_date=end_date, date_type=MagDateType.archived)
        self.assertEqual(query, "ArchivedDate lt datetime'2020-05-01T00:00Z'")

        query = make_mag_query(end_date=end_date, date_type=MagDateType.done)
        self.assertEqual(query, "DoneDate lt datetime'2020-05-01T00:00Z'")

        # Start date, end date and date type
        query = make_mag_query(start_date=start_date, end_date=end_date, date_type=MagDateType.release)
        self.assertEqual(query, "ReleaseDate ge datetime'2020-04-01T00:00Z' and ReleaseDate lt "
                                "datetime'2020-05-01T00:00Z'")

        query = make_mag_query(start_date=start_date, end_date=end_date, date_type=MagDateType.discovered)
        self.assertEqual(query, "DiscoveredDate ge datetime'2020-04-01T00:00Z' and DiscoveredDate lt "
                                "datetime'2020-05-01T00:00Z'")

        query = make_mag_query(start_date=start_date, end_date=end_date, date_type=MagDateType.archived)
        self.assertEqual(query, "ArchivedDate ge datetime'2020-04-01T00:00Z' and ArchivedDate lt "
                                "datetime'2020-05-01T00:00Z'")

        query = make_mag_query(start_date=start_date, end_date=end_date, date_type=MagDateType.done)
        self.assertEqual(query, "DoneDate ge datetime'2020-04-01T00:00Z' and DoneDate lt "
                                "datetime'2020-05-01T00:00Z'")

        # State, start date, end date and date type
        query = make_mag_query(state=MagState.discovered, start_date=start_date, end_date=end_date,
                               date_type=MagDateType.discovered)
        self.assertEqual(query, "State eq 'discovered' and DiscoveredDate ge datetime'2020-04-01T00:00Z' "
                                "and DiscoveredDate lt datetime'2020-05-01T00:00Z'")

        query = make_mag_query(state=MagState.archived, start_date=start_date, end_date=end_date,
                               date_type=MagDateType.archived)
        self.assertEqual(query, "State eq 'archived' and ArchivedDate ge datetime'2020-04-01T00:00Z' "
                                "and ArchivedDate lt datetime'2020-05-01T00:00Z'")

        query = make_mag_query(state=MagState.done, start_date=start_date, end_date=end_date,
                               date_type=MagDateType.done)
        self.assertEqual(query, "State eq 'done' and DoneDate ge datetime'2020-04-01T00:00Z' "
                                "and DoneDate lt datetime'2020-05-01T00:00Z'")


def make_mag_release(account_name: str, account_key: str, year: int, month: int, day: int):
    min_date = pendulum.datetime(1601, 1, 1)
    partition_key_ = 'mag'
    row_key_ = f'mag-{year:0>4d}-{month:0>2d}-{day:0>2d}'
    state_ = MagState.discovered
    task_ = MagTask.not_started
    release_date_ = pendulum.datetime(year=year, month=month, day=day)
    source_container_ = row_key_
    source_container_last_modified_ = pendulum.datetime(year=year, month=month, day=day, hour=1)
    release_container_ = ''
    release_path_ = ''
    discovered_date_ = pendulum.datetime(year=year, month=month, day=day, hour=2)
    archived_date_ = min_date
    done_date_ = min_date
    return MagRelease(partition_key_, row_key_, state_, task_, release_date_, source_container_,
                      source_container_last_modified_, release_container_, release_path_, discovered_date_,
                      archived_date_, done_date_, account_name=account_name, account_key=account_key)


class TestMagRelease(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestMagRelease, self).__init__(*args, **kwargs)
        self.account_name = os.getenv('STORAGE_ACCOUNT_NAME')
        self.account_key = os.getenv('STORAGE_ACCOUNT_KEY')
        create_table(self.account_name, self.account_key, MagRelease.TABLE_NAME)

    def test_secrets_hidden(self):
        # Check that account key is hidden
        account_name = 'myaccountname'
        secret = 'secret'

        # Check that account_key and sas_token are hidden
        release = make_mag_release(account_name, secret, 2020, 1, 1)
        self.assertIn('account_key=hidden', release.__repr__())
        self.assertNotIn(secret, release.__str__())
        self.assertNotIn(secret, release.__repr__())

        # Check that account_key is None
        release = make_mag_release(account_name, None, 2020, 1, 1)
        self.assertIn('account_key=None', release.__repr__())

    def test_create(self):
        release = make_mag_release(self.account_name, self.account_key, 2019, 6, 1)
        try:
            success = release.create()
            self.assertTrue(success)
        finally:
            release.delete()

    def test_delete(self):
        release = make_mag_release(self.account_name, self.account_key, 2019, 6, 1)

        # Check that we can create and then delete
        release.create()
        release.delete()

        # Check that second delete fails
        with self.assertRaises(AzureMissingResourceHttpError):
            release.delete()

    def test_update(self):
        release = make_mag_release(self.account_name, self.account_key, 2019, 6, 1)
        try:
            release.create()

            # Update release
            release.state = MagState.archived
            release.archived_date = pendulum.utcnow().microsecond_(0)
            release.update()

            # Verify that release is updated
            service = TableService(account_name=self.account_name, account_key=self.account_key)
            entity = service.get_entity(MagRelease.TABLE_NAME, release.partition_key, release.row_key)
            updated_release = MagRelease.from_entity(entity)
            self.assertEqual(release.state, updated_release.state)
            self.assertEqual(release.archived_date, updated_release.archived_date)
        finally:
            release.delete()


def make_containers():
    containers = []
    cp1 = ContainerProperties()
    cp1.name = 'mag-2020-04-17'
    cp1.last_modified = pendulum.datetime(year=2020, month=4, day=18)
    containers.append(cp1)

    cp3 = ContainerProperties()
    cp3.name = 'mag-2020-05-01'
    cp3.last_modified = pendulum.datetime(year=2020, month=5, day=1)
    containers.append(cp3)

    cp2 = ContainerProperties()
    cp2.name = 'mag-2020-04-24'
    cp2.last_modified = pendulum.datetime(year=2020, month=4, day=25)
    containers.append(cp2)

    return containers


class TestMagArchiverClient(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestMagArchiverClient, self).__init__(*args, **kwargs)
        self.account_name = os.getenv('STORAGE_ACCOUNT_NAME')
        self.account_key = os.getenv('STORAGE_ACCOUNT_KEY')
        create_table(self.account_name, self.account_key, MagRelease.TABLE_NAME)

    def test_secrets_hidden(self):
        # Check that account key is hidden
        account_name = 'myaccountname'
        secret = 'secret'

        # Check that account_key and sas_token are hidden
        client = MagArchiverClient(account_name=account_name, account_key=secret, sas_token=secret)
        expected = f'MagArchiverClient(account_name={account_name}, account_key=hidden, sas_token=hidden)'
        self.assertEqual(client.__str__(), expected)
        self.assertEqual(client.__repr__(), expected)
        self.assertNotIn(secret, client.__str__())
        self.assertNotIn(secret, client.__repr__())

        # Check that account_key and sas_token are None
        client = MagArchiverClient(account_name=account_name)
        expected = f'MagArchiverClient(account_name={account_name}, account_key=None, sas_token=None)'
        self.assertEqual(client.__str__(), expected)
        self.assertEqual(client.__repr__(), expected)

    @patch('mag_archiver.mag.list_containers')
    @patch('pendulum.datetime.now')
    def test_list_containers(self, mock_now, mock_list_containers):
        # Mock time
        mock_now.return_value = pendulum.datetime(year=2020, month=5, day=1, minute=10)

        # Mock containers
        containers_in = make_containers()
        mock_list_containers.return_value = containers_in

        # Test that 2 containers are returned when last_modified_thresh=1
        client = MagArchiverClient(account_name=self.account_name, account_key=self.account_key)
        containers_out = client.list_containers(last_modified_thresh=1)
        self.assertEqual(len(containers_out), 2)

        # Test that 3 containers are returned when last_modified_thresh=0
        containers_out = client.list_containers(last_modified_thresh=0)
        self.assertEqual(len(containers_out), 3)

        # Test sort order reverse=False
        self.assertEqual(containers_in[0].name, containers_out[0].name)
        self.assertEqual(containers_in[2].name, containers_out[1].name)
        self.assertEqual(containers_in[1].name, containers_out[2].name)

        # Test sort order reverse=True
        containers_out = client.list_containers(last_modified_thresh=0, reverse=True)
        self.assertEqual(len(containers_out), 3)
        self.assertEqual(containers_in[1].name, containers_out[0].name)
        self.assertEqual(containers_in[2].name, containers_out[1].name)
        self.assertEqual(containers_in[0].name, containers_out[2].name)

    @patch('mag_archiver.mag.list_containers')
    @patch('pendulum.datetime.now')
    def test_update_releases(self, mock_now, mock_list_containers):
        # Mock time
        mock_now.return_value = pendulum.datetime(year=2020, month=5, day=1, minute=10)

        # Mock containers
        containers_in = make_containers()
        mock_list_containers.return_value = containers_in

        # Mock fetching of containers
        client = MagArchiverClient(account_name=self.account_name, account_key=self.account_key)
        containers = client.list_containers(last_modified_thresh=1)

        try:
            # Update releases based on containers
            num_updated, num_errors = client.update_releases(containers)
            self.assertEqual(num_updated, 2)
            self.assertEqual(num_errors, 0)
        finally:
            # Clean up
            service = TableService(account_name=self.account_name, account_key=self.account_key)
            for container in containers:
                service.delete_entity(MagRelease.TABLE_NAME, 'mag', container.name.replace("mag-", ""))

    @patch('mag_archiver.mag.list_containers')
    @patch('pendulum.datetime.now')
    def test_list_releases(self, mock_now, mock_list_containers):
        # Mock time
        mock_now.return_value = pendulum.datetime(year=2020, month=5, day=1, hour=1)

        # Mock containers
        containers_in = make_containers()
        mock_list_containers.return_value = containers_in

        # Mock fetching of containers
        client = MagArchiverClient(account_name=self.account_name, account_key=self.account_key)
        containers = client.list_containers(last_modified_thresh=1)

        try:
            # Update releases based on containers
            num_updated, num_errors = client.update_releases(containers)
            self.assertEqual(num_updated, 3)
            self.assertEqual(num_errors, 0)

            # Two releases
            start_date = pendulum.datetime(year=2020, month=4, day=17)
            end_date = pendulum.datetime(year=2020, month=5, day=1)
            releases = client.list_releases(start_date=start_date, end_date=end_date, state=MagState.discovered,
                                            date_type=MagDateType.release)
            self.assertEqual(len(releases), 2)

            # 1 release
            start_date = pendulum.datetime(year=2020, month=4, day=17, minute=1)
            end_date = pendulum.datetime(year=2020, month=5, day=1)
            releases = client.list_releases(start_date=start_date, end_date=end_date, state=MagState.discovered,
                                            date_type=MagDateType.release)
            self.assertEqual(len(releases), 1)

            # Three releases
            start_date = pendulum.datetime(year=2020, month=4, day=17)
            end_date = pendulum.datetime(year=2020, month=5, day=1, minute=1)
            releases = client.list_releases(start_date=start_date, end_date=end_date, state=MagState.discovered,
                                            date_type=MagDateType.release, reverse=False)
            self.assertEqual(len(releases), 3)

            # Sorting reverse=False
            self.assertEqual(releases[0].row_key, '2020-04-17')
            self.assertEqual(releases[1].row_key, '2020-04-24')
            self.assertEqual(releases[2].row_key, '2020-05-01')

            # Sorting reverse=True
            releases = client.list_releases(start_date=start_date, end_date=end_date,
                                            state=MagState.discovered, date_type=MagDateType.release,
                                            reverse=True)
            self.assertEqual(releases[0].row_key, '2020-05-01')
            self.assertEqual(releases[1].row_key, '2020-04-24')
            self.assertEqual(releases[2].row_key, '2020-04-17')

        finally:
            # Clean up
            service = TableService(account_name=self.account_name, account_key=self.account_key)
            for container in containers:
                service.delete_entity(MagRelease.TABLE_NAME, 'mag', container.name.replace("mag-", ""))



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import logging
import re
from enum import Enum
from typing import Optional, List, Any

import pendulum
from azure.common import AzureConflictHttpError
from azure.cosmosdb.table.models import EntityProperty, EdmType
from azure.cosmosdb.table.tableservice import TableService
from azure.storage.blob import ContainerProperties
from pendulum import Pendulum

from mag_archiver.azure import copy_container, delete_container
from mag_archiver.azure import list_containers


class MagState(Enum):
    """ The state of the MAG release """

    discovered = 'discovered'
    archived = 'archived'
    done = 'done'


class MagTask(Enum):
    """ The current task being executed """

    not_started = 'not-started'
    copying_to_vm = 'copying-to-vm'
    archiving = 'archiving'
    copying_to_release_container = 'copying-to-release-container'
    cleaning_up = 'cleaning-up'
    done = 'done'


class MagDateType(Enum):
    """ The date field type """

    release = 'ReleaseDate'
    discovered = 'DiscoveredDate'
    archived = 'ArchivedDate'
    done = 'DoneDate'

    @classmethod
    def map(cls):
        return {
            MagDateType.release: 'release_date',
            MagDateType.discovered: 'discovered_date',
            MagDateType.archived: 'archived_date',
            MagDateType.done: 'done_date',
        }

    def attr(self):
        map_ = MagDateType.map()
        return map_[self]


def make_mag_query(start_date: Optional[Pendulum] = None, end_date: Optional[Pendulum] = None,
                   state: Optional[MagState] = None, date_type: MagDateType = MagDateType.release) -> str:
    """ Build a query for the MagReleases table.

    :param start_date: start date for the query (inclusive).
    :param end_date: end date for the query (exclusive).
    :param state: the state of the MAG release.
    :param date_type: the date type to query.
    :return: the query.
    """

    commands = []
    date_format = '%Y-%m-%dT%H:%MZ'

    if state is not None:
        commands.append(f"State eq '{state.value}'")

    if start_date is not None:
        commands.append(f"{date_type.value} ge datetime'{start_date.strftime(date_format)}'")

    if end_date is not None:
        commands.append(f"{date_type.value} lt datetime'{end_date.strftime(date_format)}'")

    query = ' and '.join(commands)
    return query


def hide_if_not_none(secret: Any):
    """ Hide a secret unless it is None, which means that the user didn't set it.

    :param secret: the secret.
    :return: None or 'hidden'
    """

    value = None
    if secret is not None:
        value = 'hidden'
    return value


class MagContainer:

    def __init__(self, name: str, last_modified: Pendulum, release_date: Pendulum):
        """ Microsoft Academic Graph container class.

        :param name: name of the container.
        :param last_modified: date that the container was last modified.
        :param release_date: date that the MAG release was released.
        """

        self.name = name
        self.last_modified = last_modified
        self.release_date = release_date

    def __str__(self):
        return f'Mag Container {self.name}'

    def __repr__(self):
        return f'MagContainer({self.name}, {self.last_modified.strftime("%Y-%m-%d")}, ' \
               f'{self.release_date.strftime("%Y-%m-%d")})'


class MagRelease:
    TABLE_NAME = 'MagReleases'
    __PARTITION_KEY = 'PartitionKey'
    __ROW_KEY = 'RowKey'
    __STATE = 'State'
    __TASK = 'Task'
    __RELEASE_DATE = 'ReleaseDate'
    __SOURCE_CONTAINER = 'SourceContainer'
    __SOURCE_CONTAINER_LAST_MODIFIED = 'SourceContainerLastModified'
    __RELEASE_CONTAINER = 'ReleaseContainer'
    __RELEASE_PATH = 'ReleasePath'
    __DISCOVERED_DATE = 'DiscoveredDate'
    __ARCHIVED_DATE = 'ArchivedDate'
    __DONE_DATE = 'DoneDate'

    def __init__(self, partition_key: str, row_key: str, state: MagState, task: MagTask, release_date: Pendulum,
                 source_container: str, source_container_last_modified: Pendulum, release_container: str,
                 release_path: str, discovered_date: Pendulum, archived_date: Pendulum, done_date: Pendulum,
                 account_name: Optional[str] = None, account_key: Optional[str] = None):
        """ Microsoft Academic Graph release class.

        :param partition_key: partition key.
        :param row_key: the row key. partition_key + row_key form the primary key.
        :param state: state that the MAG release is in.
        :param task: current task that is being executed.
        :param release_date: the official release date for the MAG release. Note that releases are provisioned onto
        the Azure storage account about a week after the "release_date".
        :param source_container: container where the MAG release was provisioned to.
        :param source_container_last_modified: date that the source container was last modified.
        :param release_container: container where the MAG releases are staged for data transfer.
        :param release_path: path to the MAG release on the release container.
        :param discovered_date: date that the MAG release was discovered by the MAG archiver.
        :param archived_date: date the the MAG release was archived. Set to the minimum date, i.e.
        1601-01-01T00:00:00.000Z if not specified.
        :param done_date: date the MAG release is ready to be transferred. Set to the minimum date, i.e.
        1601-01-01T00:00:00.000Z if not specified.
        :param account_name: Azure Storage account name.
        :param account_key: Azure Storage account key.
        """

        self.partition_key = partition_key
        self.row_key = row_key
        self.state = state
        self.task = task
        self.release_date = release_date
        self.source_container = source_container
        self.source_container_last_modified = source_container_last_modified
        self.release_container = release_container
        self.release_path = release_path
        self.discovered_date = discovered_date
        self.archived_date = archived_date
        self.done_date = done_date
        self.account_name = account_name
        self.account_key = account_key

    def __assert_account(self, msg):
        assert self.account_name is not None and self.account_key is not None, msg

    def __table_service(self):
        return TableService(account_name=self.account_name, account_key=self.account_key)

    def create(self) -> bool:
        """ Create the MagRelease instance in the MagReleases table.

        :return: whether the instance was created or not.
        """

        self.__assert_account("MagRelease.create: account_name and account_key must be supplied.")
        service = self.__table_service()
        success = True
        try:
            service.insert_entity(MagRelease.TABLE_NAME, self.to_entity())
        except AzureConflictHttpError as e:
            success = False
            logging.error(e)

        return success

    def delete(self) -> None:
        """ Delete this MagRelease instance from the MagReleases table.

        :return: None.
        """

        self.__assert_account("MagRelease.delete: account_name and account_key must be supplied.")
        service = self.__table_service()
        service.delete_entity(MagRelease.TABLE_NAME, self.partition_key, self.row_key)

    def archive(self, target_container: str, target_folder: str):
        """ Archive this MAG release into to the target container.

        TODO: in the future this method should compress the release before copying it.

        :param target_container: the container to copy the MAG release too.
        :param target_folder: the folder within the target container to copy the MAG release into.
        :return: None.
        """

        self.__assert_account("MagRelease.archive: account_name and account_key must be supplied.")
        return copy_container(self.account_name, self.account_key, self.source_container, target_container,
                              target_folder)

    def cleanup(self):
        """ Cleanup the source container.

        :return:
        """

        self.__assert_account("MagRelease.cleanup: account_name and account_key must be supplied.")
        return delete_container(self.account_name, self.account_key, self.source_container)

    def update(self):
        """ Update this MAG release instance in the MagReleases table.

        :return:
        """

        self.__assert_account("MagRelease.update_state: account_name and account_key must be supplied.")

        # Update properties
        service = self.__table_service()
        return service.update_entity(MagRelease.TABLE_NAME, self.to_entity())

    @staticmethod
    def from_entity(entity: dict, account_name: Optional[str] = None, account_key: Optional[str] = None):
        """ Create a MagRelease instance from an entity dictionary.

        :param entity: the entity returned from the Azure Table Storage API.
        :param account_name: Azure Storage account name. Optional, only required if the create, delete, archive,
        cleanup or update functions need to be used.
        :param account_key: Azure Storage account key. Optional, only required if the create, delete, archive,
        cleanup or update functions need to be used.
        :return: the MagRelease instance.
        """

        partition_key_ = entity[MagRelease.__PARTITION_KEY]
        row_key_ = entity[MagRelease.__ROW_KEY]
        state_ = MagState(entity[MagRelease.__STATE])
        task_ = MagTask(entity[MagRelease.__TASK])
        release_date_ = pendulum.instance(entity[MagRelease.__RELEASE_DATE])
        source_container_ = entity[MagRelease.__SOURCE_CONTAINER]
        source_container_last_modified_ = entity[MagRelease.__SOURCE_CONTAINER_LAST_MODIFIED]
        release_container_ = entity[MagRelease.__RELEASE_CONTAINER]
        release_path_ = entity[MagRelease.__RELEASE_PATH]
        discovered_date_ = pendulum.instance(entity[MagRelease.__DISCOVERED_DATE])
        archived_date_ = pendulum.instance(entity[MagRelease.__ARCHIVED_DATE])
        done_date_ = pendulum.instance(entity[MagRelease.__DONE_DATE])

        return MagRelease(partition_key_, row_key_, state_, task_, release_date_, source_container_,
                          source_container_last_modified_, release_container_, release_path_, discovered_date_,
                          archived_date_, done_date_, account_name=account_name, account_key=account_key)

    def to_entity(self) -> dict:
        """ Convert a MagRelease instance into an Azure Table Storage entity.

        :return: an Azure Table Storage entity.
        """

        entity = dict()
        entity[MagRelease.__PARTITION_KEY] = self.partition_key
        entity[MagRelease.__ROW_KEY] = self.row_key
        entity[MagRelease.__STATE] = self.state.value
        entity[MagRelease.__TASK] = self.task.value
        entity[MagRelease.__RELEASE_DATE] = EntityProperty(EdmType.DATETIME, value=self.release_date)
        entity[MagRelease.__SOURCE_CONTAINER] = self.source_container
        entity[MagRelease.__SOURCE_CONTAINER_LAST_MODIFIED] = \
            EntityProperty(EdmType.DATETIME, value=self.source_container_last_modified)
        entity[MagRelease.__RELEASE_CONTAINER] = self.release_container
        entity[MagRelease.__RELEASE_PATH] = self.release_path
        entity[MagRelease.__DISCOVERED_DATE] = EntityProperty(EdmType.DATETIME, value=self.discovered_date)
        entity[MagRelease.__ARCHIVED_DATE] = EntityProperty(EdmType.DATETIME, value=self.archived_date)
        entity[MagRelease.__DONE_DATE] = EntityProperty(EdmType.DATETIME, value=self.done_date)
        return entity

    def __str__(self):
        return f'MagRelease {self.release_date.strftime("%Y-%m-%d")}'

    def __repr__(self):
        dt_format = '%Y-%m-%dT%H:%MZ'

        return 'MagRelease(' \
               f'{self.partition_key}, ' \
               f'{self.row_key}, ' \
               f'{self.state.value}, ' \
               f'{self.task.value}, ' \
               f'{self.release_date.strftime(dt_format)}, ' \
               f'{self.source_container}, ' \
               f'{self.source_container_last_modified.strftime(dt_format)}, ' \
               f'{self.release_container}, ' \
               f'{self.release_path}, ' \
               f'{self.discovered_date.strftime(dt_format)}, ' \
               f'{self.archived_date.strftime(dt_format)}, ' \
               f'{self.done_date.strftime(dt_format)}, ' \
               f'account_name={self.account_name}, ' \
               f'account_key={hide_if_not_none(self.account_key)})'


class MagArchiverClient:
    __MAG_RELEASE_RE = re.compile("mag-[0-9]{4}-[0-9]{2}-[0-9]{2}")

    def __init__(self, account_name: Optional[str] = None, account_key: Optional[str] = None,
                 sas_token: Optional[str] = None):
        """

        :param account_name: the name of the Azure Storage account where the Microsoft Academic Graph releases are
        provisioned and the `MagReleases` Azure Storage Table is stored.
        :param account_key: the key to Azure Storage account.
        :param sas_token: the shared access token to use for authentication instead of the account_key.
        """

        self.account_name = account_name
        self.account_key = account_key
        self.sas_token = sas_token

    def list_containers(self, last_modified_thresh: Optional[float] = None, reverse: bool = False) \
            -> List[MagContainer]:
        """ List all Azure Storage Blob Containers holding MAG releases.

        :param last_modified_thresh: only include containers that were last modified greater than or equal to a
        specific number of hours.
        :param reverse: sort from oldest to newest using release_date datetime.
        :return: the containers holding MAG releases.
        """

        # List all containers in the storage account
        containers: List[ContainerProperties] = list_containers(self.account_name, self.account_key)

        tz = pendulum.timezone('UTC')
        current_time = pendulum.datetime.now(tz)

        # Select all containers containing MAG releases that were last updated greater than
        mag_containers = []
        for container in containers:
            container_name = container.name
            last_modified = pendulum.instance(container.last_modified, tz=tz)
            hours_since_modified = current_time.diff(last_modified).in_hours()

            # Only include containers with a name that matches the MAG container pattern
            if MagArchiverClient.__MAG_RELEASE_RE.match(container_name) is not None:
                # Add of no last_modified_thresh supplied or if the hours since modified is greater than the last
                # modified threshold
                if last_modified_thresh is None or last_modified_thresh <= hours_since_modified:
                    release_date_str = container.name.replace("mag-", "")
                    release_date = pendulum.parse(release_date_str)
                    mag_container = MagContainer(container_name, last_modified, release_date)
                    mag_containers.append(mag_container)

        # Sort from oldest to newest
        mag_containers.sort(key=lambda c: c.release_date, reverse=reverse)
        return mag_containers

    def update_releases(self, containers: List[MagContainer]):
        """ Update the releases in the MagReleases table based on a list of containers containing MAG releases.

        :param containers: the containers containing MAG releases.
        :return: the number of releases created and the number of errors.
        """

        min_date = pendulum.datetime(1601, 1, 1)
        discovered_date = pendulum.now()

        # Get all entities
        table_service = TableService(account_name=self.account_name, account_key=self.account_key)
        entities = table_service.query_entities(MagRelease.TABLE_NAME)

        # Get all containers that are not in the MagReleases table
        new_container_index = dict.fromkeys(set([container.name for container in containers]) -
                                            set([entity['SourceContainer'] for entity in entities]), 0)
        num_new_containers = len(new_container_index)
        logging.info(f"Num new containers discovered: {num_new_containers}")

        # Only add new containers
        num_created = 0
        num_errors = 0
        for container in containers:
            if container.name in new_container_index:
                partition_key = 'mag'
                row_key = container.release_date.strftime("%Y-%m-%d")
                release = MagRelease(partition_key, row_key, MagState.discovered, MagTask.not_started,
                                     container.release_date, container.name, container.last_modified, '', '',
                                     discovered_date, min_date, min_date, account_name=self.account_name,
                                     account_key=self.account_key)
                success = release.create()
                if success:
                    num_created += 1
                else:
                    num_errors += 1
        return num_created, num_errors

    def list_releases(self, start_date: Optional[Pendulum] = None,
                      end_date: Optional[Pendulum] = None, state: Optional[MagState] = None,
                      date_type: MagDateType = MagDateType.release, reverse: bool = False) -> List[MagRelease]:
        """ List Microsoft Academic releases in the MagReleases Azure Storage Table.

        :param start_date: start date for the query.
        :param end_date: end date for the query.
        :param state: the state of the MAG release.
        :param date_type: the date type to query.
        :param reverse: whether to reverse sort the results. Sorting is performed on the date type.
        :return: the list of MagReleases found.
        """

        # Query and fetch releases
        table_service = TableService(account_name=self.account_name, account_key=self.account_key,
                                     sas_token=self.sas_token)
        query = make_mag_query(start_date=start_date, end_date=end_date, state=state, date_type=date_type)
        entities = table_service.query_entities('MagReleases', filter=query)

        # Convert entities into MagRelease objects
        releases = []
        for entity in entities:
            release = MagRelease.from_entity(entity, account_name=self.account_name, account_key=self.account_key)
            releases.append(release)

        # Sort from oldest to newest, unless reverse is set
        releases.sort(key=lambda r: getattr(r, MagDateType.attr(date_type)), reverse=reverse)
        return releases

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return 'MagArchiverClient(' \
               f'account_name={self.account_name}, ' \
               f'account_key={hide_if_not_none(self.account_key)}, ' \
               f'sas_token={hide_if_not_none(self.sas_token)})'



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import logging
import os
from typing import List

import azure.functions as func
import pendulum

from mag_archiver.mag import MagArchiverClient, MagState, MagDateType, MagRelease, MagTask


def main(timer: func.TimerRequest) -> None:
    logging.info('mag_archiver function: started')

    # Get environment variables
    account_name = os.getenv('STORAGE_ACCOUNT_NAME')
    account_key = os.getenv('STORAGE_ACCOUNT_KEY')
    target_container = os.getenv('TARGET_CONTAINER')

    # Check all environment variables are specified
    req_env_vars = account_name is not None and account_key is not None and target_container is not None
    req_env_vars_msg = "The environment variables STORAGE_ACCOUNT_NAME, STORAGE_ACCOUNT_KEY and TARGET_CONTAINER " \
                       "must be set."
    if not req_env_vars:
        logging.error(req_env_vars_msg)
    assert req_env_vars, req_env_vars_msg

    # List MAG containers in storage account
    client = MagArchiverClient(account_name=account_name, account_key=account_key)
    containers = client.list_containers(last_modified_thresh=1)
    logging.info(f'{client}')
    logging.info(f'Discovered containers: {containers}')

    # Update MagReleases Table based on discovered containers
    num_created, num_errors = client.update_releases(containers)
    logging.info(f'Update MagReleases table: num_created={num_created}, num_errors={num_errors}')

    # List all discovered MAG releases and sort from oldest to newest based on release date
    releases: List[MagRelease] = client.list_releases(state=MagState.discovered, date_type=MagDateType.release,
                                                      reverse=False)
    num_releases = len(releases)
    logging.info(f'Number of discovered MAG releases: {num_releases}')
    logging.info(f'Discovered MAG releases: {releases}')

    # If 1 or more MAG releases was found then process the oldest one
    if len(releases) >= 1:
        release: MagRelease = releases[0]
        # We want to have the same directory structure here, as this is how the data will be structured in the
        # Google Cloud Storage bucket when the data is transferred with the Google Cloud data transfer service
        target_folder = f'telescopes/mag/{release.source_container}'
        logging.info(f'Processing: {release}')

        # Update task to copying
        release.task = MagTask.copying_to_release_container
        logging.info(f'Update task to copying: {release}')
        release.update()
        logging.info(f'Update task to copying finished: {release}')

        # Copy source container to target container
        logging.info(f'Archive: {release}')
        release.archive(target_container, target_folder)  # TODO: how do you know if the copying failed?
        logging.info(f'Archive finished: {release}')

        # Update state to archived
        #   - set state to archived
        #   - add release container and release path
        #   - set archived date to now
        #   - set task to cleaning up
        release.state = MagState.archived
        release.release_container = target_container
        release.release_path = target_folder
        release.archived_date = pendulum.utcnow()
        release.task = MagTask.cleaning_up
        logging.info(f'Update state to archived: {release}')
        release.update()
        logging.info(f'Update state to archived finished: {release}')

        # Cleanup: delete source container
        logging.info(f'Cleanup: {release}')
        release.cleanup()
        logging.info(f'Cleanup finished: {release}')

        # Update state to done
        # - set state and task to done
        # - set done date to now
        release.state = MagState.done
        release.task = MagTask.done
        release.done_date = pendulum.utcnow()
        logging.info(f'Update state to done: {release}')
        release.update()
        logging.info(f'Update state to done finished: {release}')
    else:
        logging.info(f'No discovered releases to process')

    logging.info('mag_archiver function: finished')
